(*
   module Interface (.mli file only)
   Signatures for syntax and environments. Most files will open Types instead of this.
*)

open Lolevel.Error

type identifier = string
  
  (*
     Our interpreters will be (VersionA) a substitute-and-repeat
     interpreter, and (VersionB) an environment-based interpreter. We
     discuss the differences between these in the course notes.
  
     The implementeations we provide make both of these call-by-value.
     When given a term `App(head, arg)`, the steps are: first, reduce
     or evaluate `head`---it might involve further `App`s itself.
     Second, reduce or evaluate `arg`.  Third, only _when_ `arg`
     reduces or evaluates to a result value, then "perform" the
     application. What this last step amounts to is different in the
     two interpreters.  Call-by-name interpreters would "perform" the
     application regardless, and without even trying to reduce or evaluate
     arg to a result value beforehand.
  
     Additionally, these implementations assume that free variables
     are "stuck" terms rather than result values. Result values
     include functions (whether primitives or (in VA) Lambdas or (in VB)
     Closures) and literals.
  *)

module type SYNTAX = sig

  type bare_term =
    | TermVar of identifier
    | TermApp of term * term
    | TermLambda of identifier * term
    | TermLet of identifier * term * term
    | TermLetRec of (identifier * term) list * term
    | TermIf of term * term * term
    (* Constants, functional and otherwise *)
    | TermPrimfun of primfun
    | TermConstructor of constructor
    | TermDestructor of destructor
    | TermLiteral of literal
  
  (*
     In VA/substitute-and-repeat, `result` is simply a proper subset
     of terms. (Ignoring the conversion from TermVar to Symbol.)
  
     In VB/env-based, `result` will hold `Closure`s instead of
     Lambdas. These are datatypes that associate the abstract with an
     environment that holds bindings for its free variables.
  *)
  and bare_result =
    (* TermVar ~~> Symbol when declared `symbol` at toplevel *)
    | Symbol of identifier
    | Closure of identifier * term * tmp_env (* only used as a result in VB *)
    | ResLambda of identifier * term (* only used as a result in VA *)
    | ResPrimfun of primfun
    | ResConstructor of constructor
    | ResDestructor of destructor
    | ResLiteral of literal
  
  (* Terms/results are annotated with info about where in the source
     file they originated. *)
  and term = info * bare_term
  and result = info * bare_result
  
  (* Types for primitive functions, constructors, and literals. *)

  and primfun =
    | Funct1 of (printer) * (gobbler)
    | Funct2 of (printer -> printer) * (gobbler) * (result -> gobbler)
  and constructor =
    | Cons0 of (cons_printer) * (projector)
    | Cons1 of (printer -> cons_printer) * (result -> projector) * (gobbler)
    | Cons2 of (printer -> printer -> cons_printer) * (result -> result -> projector) * (gobbler) * (result -> gobbler)
    | Cons3 of (printer -> printer -> printer -> cons_printer) * (result -> result -> result -> projector) * (gobbler) * (result -> gobbler) * (result -> result -> gobbler)
  and destructor = Fst | Snd | Trd | IsZero | Pred (* | IsNil | Head | Tail *)
  and literal = Bool of bool | Num of int | Str of string
  and cons_printer = projector option -> unit
  and projector = destructor -> result option
  and gobbler = result -> result option

  (* Abstract type, implementation provided elsewhere. *)
  and tmp_env
  
  (* Input is parsed into a list of terms, symbol declarations, and
     toplevel let/letrec bindings --- but letrec not implemented yet;
     that's for later. *)
  type toplevel_syntax =
    | Term of info * term
    | BindTerm of info * identifier * term
    | BindRec of info (* letrec not implemented *)
    | BindSymbol of info * identifier
  
  (*
     For now, the identifier "x" is bound to the same type of data
     (indeed, the same data) that `Var "x"` expresses, but in
     versions of the interpreter we explore in later weeks, that
     won't always be true.
  *)
  type bound = result
  
  (* Wrappers around what we save in environments. *)
  type binding =
    (* not used in VA/substitute-and-repeat interpreter *)
    | LocalBind of bound
    (* when `let x = ... end` was declared at toplevel *)
    | TopBind of bound
    (* when `symbol x` was declared at toplevel *)
    | SymbolBind

end (* SYNTAX *)

  (*
     The file `types.ml` provides several different implementations
     of the ENV signature declared here. You can choose between Env1
     and Env2 without affecting the program's behavior.

     The environments play only a limited role in the VA interpreter,
     and can be dispensed with entirely (select the Env0
     implementation of environments), if you forego toplevel bindings
     and `symbol` declarations.
  
     In the VB interpreter, the environments are essential (and
     Env0 won't be adequate).
  *)
module type ENV = sig
  type env
  (* The following line is needed to unify the `env` type provided by
     an implementation of the ENV signature with the corresponding
     type used in SYNTAX, above.  See also types.ml, where even more
     trickery needs to be deployed. *)
  include SYNTAX with type tmp_env = env
  val empty : env
  val shift : identifier -> binding -> env -> env
  val lookup : identifier -> env -> binding option
  val version : string
end (* ENV *)
